@@ -1,9 +1,10 @@
 from fitparse.base import FitFile, FitParseError
+from fitparse.records import DataMessage
 from fitparse.processors import FitFileDataProcessor, StandardUnitsDataProcessor
 
 
-__version__ = '1.0.0'
+__version__ = '1.2.0'
 __all__ = [
     'FitFileDataProcessor', 'FitFile', 'FitParseError',
-    'StandardUnitsDataProcessor',
+    'StandardUnitsDataProcessor', 'DataMessage'
 ]
